module user_output_column

  use binas,      only : ae, twopi, grav
  use dims,       only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend
  use dims,       only : nregions, region_name, itaur, xcyc
  use chem_param, only : ntrace, ntracet, names, fscale
  use file_hdf,   only : THdfFile, Init, Done, TSds, SetDim

  implicit none

  private
  
  public :: evaluate_columnconc, read_columnlist, stat, columns
  public :: init_column_output, reset_columnconc_accumulator
  public :: write_columnfiles, columnlistFFilename , output_columnconc
  public :: close_columnfiles, write_columnconc
  public :: n_stat, io_nrecords

  ! MODIFIED by Peter Beramaschi.
  ! PARALLEL version Maarten Krol Jul 2003
  ! jun 2005: switch to HDF output file

  character(len=100)           :: columnlistFFilename 
  character(len=100)           :: columnOutFFilename 
  character(len=100)           :: InputDir = './'
  logical                      :: output_columns_meteo = .true.
  logical                      :: write_columnfiles = .true.
  integer                      :: cstatus

  integer,save                 :: n_stat         ! number of columns 
  ! list of species for output
  integer, dimension(ntrace)   :: column_speclist
  ! number of species for output
  integer                      :: column_nspec
  ! max number of meteo data (for output in column output file)
  integer, parameter           :: n_metdat = 8
  integer, parameter           :: n_interpol = 3  ! number of column interpolation methods
  character(len=4), dimension(n_interpol), parameter :: c_interpol = (/'_grd','_slp','_int'/)
  character(len=12),dimension(n_metdat  ), parameter :: C_metdat   =  &
               (/'Temperature', &
                 'Pressure   ',&
                 'uwind      ',&
                 'vwind      ',&
                 'windspeed  ',&
                 'winddir    ',&
                 'blh        ',&
                 'gph        '/)
  character(len=12),dimension(n_metdat  ), parameter :: U_metdat   =  &
               (/'K           ', &
                 'hPa         ', &
                 'm/s         ', &
                 'm/s         ', &
                 'm/s         ', &
                 'degrees wrtN', &
                 'm           ', &
                 'm           '/)

  type(THdfFile)               :: io_hdf          ! contains HDF output file
  integer                      :: io_record = 0
  integer,save                 :: io_nrecords 
  ! 'open' datasets of the io_hdf output file:
  type(TSds), dimension(:,:), allocatable :: Sds
  type(TSds), dimension(n_metdat)         :: Sds_meteo
  type(TSds)                              :: Sds_idate

  type columns
     !cmk character(len=7)           :: ident     ! column identifier
     character(len=13)           :: ident     ! column identifier
     character(len=50)          :: name      ! column name
     real                       :: lat       ! latitude of column
     real                       :: lon       ! longitude of column
     real                       :: height       ! height of column
     real                       :: oro       ! ground height at station
     integer                    :: index     ! index of column
     integer                    :: region 
     integer                    :: is
     integer                    :: js
     integer                    :: ls
     real,dimension(n_metdat,lm(1))   :: metdat               ! meteodata (interpolated) at columns
     real,dimension(:,:,:),pointer :: c_mean              ! spcies, interpolation
     real,dimension(:,:,:),pointer :: c_std
     integer                     :: ncount              ! for averages
  end type columns

  type(columns), dimension(:), allocatable  :: stat

contains



  subroutine free_columnsfields
    use Partools, only : myid, root
    implicit none
    integer :: i_stat
    if (myid /= root) return
    do i_stat = 1, n_stat
       deallocate(stat(i_stat)%c_mean)
       deallocate(stat(i_stat)%c_std )
    enddo
    deallocate( stat )
  end subroutine free_columnsfields



  subroutine read_columnlist
    use toolbox,      only : escape_tm
#ifdef MPI  
    use mpi_const, only: myid, root, mpi_integer, mpi_character, &
         my_real, mpi_comm_world, ierr
#else
    use ParTools, only: myid, root
#endif
    use GO,          only : TrcFile, Init, Done, ReadRc
    use global_data, only : rcfile
    

    implicit none

    ! local
    integer                           :: slen1 !string length
    character(len=200)                :: dummy
    integer                           :: fstatus 
    integer                           :: i_stat
    integer                           :: region
    integer,dimension(6)              :: idate_sim
    integer                           :: itau_sim, isim, ispec
    ! x,y resolution (in degrees) for current region 
    real                              :: dxr, dyr 
    character(len=8)                  :: name_spec    !CMK changed from 6-->8
    integer                           :: sunit0=110, iostat
    type(TrcFile)                     :: rcF

    ! start
    if ( myid == root ) then
       call Init( rcF, rcfile,cstatus)
       call ReadRc( rcF, 'input.column.filename', columnlistFFilename,cstatus)
       slen1=len_trim(columnlistFFilename) 
       call ReadRc( rcF, 'inputdir.column', InputDir,cstatus)
       call Done( rcF,cstatus)
       slen1=len_trim(columnlistFFilename) 
       if ( slen1 > 0 ) then
          !=========================================
          ! open columnFile (with list of columns)
          !=========================================
          open(UNIT=sunit0, FORM='FORMATTED', &
               FILE=trim(InputDir)//'/'//trim(columnlistFFilename), &
               STATUS='OLD', ERR=1000)      
          read(sunit0, '(a)') dummy

          ! count number of columns
          n_stat=0
          stat_loop: DO
             read(sunit0,  '(a)', END=100) dummy
             if (dummy(1:7) /=  'species') then   
                n_stat = n_stat+1 
             else
                column_nspec = 0
                do
                   read(sunit0, '(a8)' , end=100) name_spec    ! CMK changed a6-->a8
                   !CMK not allowed yet to write short-lived: TODO
                   specloop: do ispec = 1, ntracet
                      if (name_spec == names(ispec) ) then
                         column_nspec = column_nspec + 1
                         column_speclist(column_nspec) = ispec
                         exit specloop
                      endif
                   enddo specloop
                enddo
                exit stat_loop
             end if
          end do stat_loop
100       continue

          print *, '_________________________________________________' &
               //'_______________________________'
          print *, n_stat, ' columns read from file: ', columnlistFFilename
          print *, column_nspec, ' concentrations will be gathered:'
          if (column_nspec == 0) then 
             print *, 'ERROR user_output_column'
             print *, 'ERROR no species for output'
             print *, 'ERROR Set column_output to .F. or give valid species name'
             call escape_tm('Error reading column file')
          endif

          do ispec = 1, column_nspec
             print '(2i4,2x,a8)', ispec, column_speclist(ispec), &   ! CMK change a6--->a8
                  names(column_speclist(ispec))
          enddo
          print *, '_________________________________________________' &
               //'_______________________________'


          allocate(stat(n_stat))

          do i_stat = 1, n_stat
             allocate(stat(i_stat)%c_mean(column_nspec,n_interpol,lm(1)))
             allocate(stat(i_stat)%c_std(column_nspec,n_interpol,lm(1)))
          enddo
          dyr = dy/yref(1)
          dxr = dx/xref(1)

          !===============
          ! read columns
          !===============

          rewind(sunit0)
          read(sunit0, '(a)') dummy
          do i_stat=1, n_stat 

             read(sunit0,  '(a7,1x,f8.2,f8.2,f8.1,1x,a50)') &
                  stat(i_stat)%ident, &
                  stat(i_stat)%lat, &
                  stat(i_stat)%lon, &
                  stat(i_stat)%height, &
                  stat(i_stat)%name

             stat(i_stat)%oro = -99999.0

             stat(i_stat)%index = i_stat

             ! assume global region as default for average mixing ratios
             stat(i_stat)%region = 1
             stat(i_stat)%is = &
                  int((stat(i_stat)%lon-float(xbeg(1)))/dxr + 0.99999) 
             stat(i_stat)%js = &
                  int((stat(i_stat)%lat-float(ybeg(1)))/dyr + 0.99999)
          end do

          ! close columnfile
          close(sunit0)
          do i_stat=1, n_stat
            do region=1, nregions
      
              dyr = dy/yref(region)
              dxr = dx/xref(region)

              if ( (stat(i_stat)%lon .ge. xbeg(region) .and. &
                    stat(i_stat)%lon .le. xend(region)) .and.      &
                   (stat(i_stat)%lat .ge. ybeg(region) .and. &
                    stat(i_stat)%lat .le. yend(region) ) ) then
                   !=====================
                   ! column is in region
                   !=====================

                ! determine region with hightes refinement factor
                if (xref(region) > xref(stat(i_stat)%region)) then
                  stat(i_stat)%region = region
                  stat(i_stat)%is = &
                    int((stat(i_stat)%lon-float(xbeg(region)))/dxr + 0.99999) 
                  stat(i_stat)%js = &
                    int((stat(i_stat)%lat-float(ybeg(region)))/dyr + 0.99999)
                end if
              end if
            end do
          end do
          do i_stat=1, n_stat 
             print '(a7,1x,f8.2,f8.2,f8.1,1x,i4,1x,a50)', &
                  stat(i_stat)%ident, &
                  stat(i_stat)%lat, &
                  stat(i_stat)%lon, &
                  stat(i_stat)%height, &
                  stat(i_stat)%region, &
                  stat(i_stat)%name
          end do
       end if
    end if  ! myid = root  
    call reset_columnconc_accumulator

    return

    ! error handling
1000 call escape_tm( 'read_columnlist: could not open '//trim(InputDir)//'/'//trim(columnlistFFilename))

  end subroutine read_columnlist


  !=====================================================================
  !=====================================================================

  subroutine init_column_output
    !=====================================
    ! HDF output for columns 
    ! (instantaneous model results)
    !=====================================
    use ParTools ,   only: myid, root
    use Dims,        only : itaui, itaue, ndyn_max
    use file_hdf,    only : Init, WriteAttribute, WriteData, TSds, Done
    use GO,          only : TrcFile, Init, Done, ReadRc
    use global_data, only : rcfile, outdir
    implicit none

    integer                           :: i_stat
    type(TrcFile)                     :: rcF
    character(len=12)                 :: xname
    integer                           :: i, ii, j, iostat, io_status
    type(TSds)                        :: sds_temp


    if (myid /= root) return
    if (write_columnfiles)then
       call Init( rcF, rcfile,cstatus)
       call ReadRc( rcF, 'output.column.filename', columnOutFFilename,cstatus)
       call Done( rcF ,cstatus)
       call Init( io_hdf, trim(OutDir)//'/'// trim(columnOutFFilename),'create' , io_status)
       io_record = 0
       io_nrecords =abs(itaue-itaui)/ndyn_max
       call WriteAttribute( io_hdf, 'io_nrecords', io_nrecords,cstatus)
       call WriteAttribute( io_hdf, 'n_stat', n_stat,cstatus)
       call WriteAttribute( io_hdf, 'n_spec', column_nspec,cstatus)
       call WriteAttribute( io_hdf, 'column_speclist', &
                       column_speclist(1:column_nspec),cstatus)
       call WriteAttribute( io_hdf, 'column_specnames', &
                        names(column_speclist(1:column_nspec)),cstatus)
       call WriteAttribute( io_hdf, 'column_lat', stat(:)%lat,cstatus)
       call WriteAttribute( io_hdf, 'column_lon', stat(:)%lon,cstatus)
       call WriteAttribute( io_hdf, 'column_height', stat(:)%height,cstatus)
       call WriteAttribute( io_hdf, 'column_index', stat(:)%index,cstatus)
       call WriteAttribute( io_hdf, 'column_region', stat(:)%region,cstatus)
       call WriteAttribute( io_hdf, 'column_is', stat(:)%is,cstatus)
       call WriteAttribute( io_hdf, 'column_js', stat(:)%js,cstatus)
       call WriteAttribute( io_hdf, 'column_names', stat(:)%name,cstatus)
       call WriteAttribute( io_hdf, 'column_ident', stat(:)%ident,cstatus)

       allocate(Sds(column_nspec, n_interpol))
       do i=1, column_nspec
          do ii= 1,n_interpol
             xname = trim(names(column_speclist(i)))//c_interpol(ii)
             call init(Sds(i,ii), io_hdf, xname, (/n_stat, lm(1),io_nrecords/), 'real(4)',cstatus)
             call SetDim( sds(i,ii), 0, 'n_columns', 'column number', (/ (j, j=1,n_stat) /) ,cstatus)
             call SetDim( sds(i,ii), 1, 'n_levels ', 'level #', (/ (j, j=1,lm(1)) /) ,cstatus)
             call SetDim( sds(i,ii), 2, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
             call WriteAttribute(Sds(i,ii), 'Unit', 'Volume mixing ratio',cstatus)
          enddo
       enddo
       do i=1,n_metdat
          call init(sds_meteo(i), io_hdf, c_metdat(i), (/n_stat, lm(1),io_nrecords/), 'real(4)',cstatus)
          call SetDim( sds_meteo(i), 0, 'n_columns', 'column number', (/ (j, j=1,n_stat) /),cstatus)
          call SetDim( sds_meteo(i), 1, 'n_levels ', 'level #', (/ (j, j=1,lm(1)) /) ,cstatus)
          call SetDim( sds_meteo(i), 2, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
          call WriteAttribute(Sds_meteo(i), 'Unit', u_metdat(i),cstatus)
       enddo
       call init(sds_idate, io_hdf, 'idate', (/6, io_nrecords/), 'integer(2)',cstatus)
       call SetDim( sds_idate, 1, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
       call WriteAttribute(Sds_idate, 'Unit', 'year month day hour minutes seconds',cstatus)
   end if

  end subroutine init_column_output

  !=========================================================================
  !=========================================================================


  subroutine close_columnfiles

    use ParTools , only: myid, root
    use file_hdf,  only: Done
    implicit none
    integer                           :: io_status, i, ii

    if (myid == root) then
       print *, myid, root, 'Write column files'
       if (write_columnfiles)then
          do i=1, column_nspec
             do ii= 1,n_interpol
                call Done(sds(i,ii),io_status)
             enddo
          enddo
          do i=1,n_metdat
             call Done(sds_meteo(i),io_status)
          enddo
          call Done(sds_idate,io_status)
          deallocate(Sds)
          call Done(io_hdf, io_status)
          print *, 'column file closed with status:', io_status
       endif
    end if   ! myid == root
    call free_columnsfields
  end subroutine close_columnfiles


  subroutine output_columnconc(region)
    !=====================================================================
    !=====================================================================

#ifdef MPI  
    use mpi_const !, only: myid, root
    use mpi_comm, only: gather_tracer_k, gather_tracer_t, stopmpi, gather_tracer_k_3d
#endif
    use ParTools, only: myid, root, root_t, root_k
    use global_data, only : mass_dat,  conv_dat
    use datetime,    only : tau2date
    use toolbox,     only : escape_tm
    use Meteo,       only : gph_dat, poro=>oro, temper_dat, pu_dat, pv_dat

    implicit none

    ! input/output
    integer,intent(in)  :: region

    ! local
    real,dimension(:,:,:,:),pointer :: rm                   ! tracer mass
    real,dimension(:,:,:,:),pointer :: rxm,rym,rzm          ! slopes of tracer mass
    real,dimension(:,:,:)  ,pointer :: m                    ! air mass
    real,dimension(:,:,:)  ,pointer :: gph                  ! geopotential height
    real,dimension(:,:)    ,pointer :: oro                  ! orography

    real, dimension(:,:,:,:), allocatable, target      :: rmg                      ! MPI arrays to gather fields.
    real, dimension(:,:,:,:), allocatable, target      :: rxmg, rymg, rzmg 
    real, dimension(:,:,:  ), allocatable, target      :: mg

    integer                         :: i_interpol           ! interpolation procedures:   
    ! 1: no interpolation
    ! 2: interpolation based on slopes
    ! 3: 3D linear interpolations


    integer                         :: i,j,l                ! grid cell indices
    integer                         :: is,js,ls             ! i,j,l index of grid cell in which column lis located
    integer                         :: isn,jsn,lsn          ! i,j,l index of grid cell which is taken as neighbour for linear interpolation
    integer                         :: lst, lstn            ! MPI implementation
    integer                         :: js_north, js_south   ! j index of grid cell for neighbours for windspeed_v interpolation
    integer                         :: is_west              ! i index of grid cell for neighbour for windspeed_u interpolation
    integer                         :: n                    ! index of tracer
    integer                         :: i_stat               ! index of column
    real                            :: ris,rjs,rls          ! deviation of column from center of the grid cell (-0.5 ... +0.5)
    real                            :: wcx,wcy,wcz          ! x,y,z-weight of grid cell (1.0 ... 0.5) 
    real                            :: dxr, dyr             ! x,y resolution (in degrees) for current region 

    real,dimension(0:lm(region))    :: height               ! height of vertical grid boundaries 
    integer, dimension(6)           :: idate_f              ! current idate for region 
    integer                         :: sunit                ! unit number for column output file
    real,dimension(lm(1))           :: rm_interpol          ! tracer mass, interpolated to column location 
    integer                         :: ispec                ! index over activated tracers

    !=============
    ! meteo output
    !=============
    real,dimension(n_stat,n_metdat,lm(1)) :: metdat               ! meteodata (interpolated) at columns
    real,dimension(:,:,:), pointer  :: T                    ! temperature
    real,dimension(:,:,:), pointer  :: phlb                 ! pressure grid boundaries
    real,dimension(:,:,:), pointer  :: pu                   ! mass flux x-direction [kg/s]
    real,dimension(:,:,:), pointer  :: pv                   ! mass flux y-direction [kg/s]
    real,dimension(:,:), pointer    :: blh                  ! boundary layer height [m]
    real,dimension(lm(1))           :: T_interpol           ! temperature, interpolated to column location 
    real,dimension(lm(1))           :: p_interpol           ! pressure, interpolated to column location 
    real,dimension(lm(1))           :: windspeed_u_interpol ! wind speed [m/s] x-direction , interpolated to column location
    real,dimension(lm(1))           :: windspeed_v_interpol ! wind speed [m/s] y-direction , interpolated to column location
    real,dimension(lm(1))           :: windspeed_interpol   ! wind speed [m/s]             , interpolated to column location
    real,dimension(lm(1))           :: winddir_interpol     ! wind direction               , interpolated to column location 
    real                            :: blh_interpol         ! boundary layer height        , interpolated to column location 
    real                            :: lat_j                ! latitude of grid cell center js  [degrees]
    real                            :: lat_jn               ! latitude of grid cell center jsn [degrees]
    real                            :: ddx_j                ! x-extension of grid cell js    [m]
    real                            :: ddx_jn               ! x-extension of grid cell jsn   [m]
    real                            :: ddy                  ! y-extension of grid cell js    [m]
    real,dimension(lm(1))           :: vmr                  ! volume mi

    integer   :: communicator,root_id,lmr,imr,jmr

    ! start

    ! assign pointers depending on the paralel status
    ! (over levels after chemistry!)
    imr = im(region) ; jmr = jm(region) ; lmr = lm(region)

#ifdef MPI

    if ( root_k /= root_t ) print *, 'Warning from MPI column output: ', &
         'root_k and root_t are not equal'

    allocate( rmg (-1:imr+2,-1:jmr+2,lmr, ntracet) ) ; rmg  = 0.0
    allocate( rxmg(-1:imr+2,-1:jmr+2,lmr, ntracet) ) ; rxmg = 0.0
    allocate( rymg(-1:imr+2,-1:jmr+2,lmr, ntracet) ) ; rymg = 0.0
    allocate( rzmg(-1:imr+2,-1:jmr+2,lmr, ntracet) ) ; rzmg = 0.0
    allocate( mg  (-1:imr+2,-1:jmr+2,lmr         ) ) ; mg   = 0.0
    if ( which_par == 'tracer' ) then 
       m => mass_dat(region)%m_t
       rm => mass_dat(region)%rm_t
       rxm => mass_dat(region)%rxm_t
       rym => mass_dat(region)%rym_t
       rzm => mass_dat(region)%rzm_t
       call gather_tracer_t(rmg,imr,jmr,lmr,2,2,0,ntracet,rm,.false.)
       call gather_tracer_t(rxmg,imr,jmr,lmr,2,2,0,ntracet,rxm,.false.)
       call gather_tracer_t(rymg,imr,jmr,lmr,2,2,0,ntracet,rym,.false.)
       call gather_tracer_t(rzmg,imr,jmr,lmr,2,2,0,ntracet,rzm,.false.)
       nullify(rm)   ! reassign pointers
       nullify(rxm)
       nullify(rym)
       nullify(rzm)
       rm => rmg
       rxm => rxmg
       rym => rymg
       rzm => rzmg
       communicator=com_trac  !WP! assign com_trac as communicator
       root_id=root_t         !root_t is the PE that knows the result.
    else
       m => mass_dat(region)%m_k
       rm => mass_dat(region)%rm_k
       rxm => mass_dat(region)%rxm_k
       rym => mass_dat(region)%rym_k
       rzm => mass_dat(region)%rzm_k
       call gather_tracer_k(rmg,imr,jmr,lmr,2,2,0,ntracet,rm,.false.)
       call gather_tracer_k(rxmg,imr,jmr,lmr,2,2,0,ntracet,rxm,.false.)
       call gather_tracer_k(rymg,imr,jmr,lmr,2,2,0,ntracet,rym,.false.)
       call gather_tracer_k(rzmg,imr,jmr,lmr,2,2,0,ntracet,rzm,.false.)
       call gather_tracer_k_3d( mg, imr,jmr,lmr, 2,2,0, m, .false.)
       
       nullify(m)   ! reassign pointers
       nullify(rm) 
       nullify(rxm)
       nullify(rym)
       nullify(rzm)
       m => mg
       rm => rmg
       rxm => rxmg
       rym => rymg
       rzm => rzmg
       communicator=com_lev  !WP! assign com_lev as communicator
       root_id=root_k
    end if
#else
    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
#endif
    ! assign pointers (METEO)
    gph => gph_dat(region)%data
    oro => poro(region)%surf
    t => temper_dat(region)%data
    pu => pu_dat(region)%data
    pv => pv_dat(region)%data
    blh => conv_dat(region)%blh
    ! point to pressure paralel over tracers (contains all layers!)
    phlb => mass_dat(region)%phlb_t

    if ( myid == root_t ) then 
       ! perform only on root: tracers are already gathered here
       ! x and y resolution [degrees] for current region

       dyr = dy/yref(region)
       dxr = dx/xref(region)

       ! idate for current region
       call tau2date(itaur(region),idate_f)


       !====================
       ! loop over columns
       !====================

       do i_stat=1, n_stat

          IF (stat(i_stat)%region == region)THEN

             !avoid edges!
             ris = (stat(i_stat)%lon-float(xbeg(region)))/dxr + 0.99999
             rjs = (stat(i_stat)%lat-float(ybeg(region)))/dyr + 0.99999
             is  = int(ris)   ! i-index of grid cell in which column is located
             js  = int(rjs)   ! j-index of grid cell in which column is located
             ! x-deviation from the center of the grid cell  (-0.5...+0.5)
             ris = ris-is-0.5
             ! y-deviation from the center of the grid cell  (-0.5...+0.5)
             rjs = rjs-js-0.5

             !=================================
             !the neighbour for x interpolation
             !=================================
             if ( ris .gt. 0 ) then 
                isn = is+1     
             else
                isn = is-1
             endif

             !=================================
             !the neighbour for y interpolation
             !=================================
             if ( rjs .gt. 0 ) then 
                jsn = js+1     
             else
                jsn = js-1
             endif

             ! x- / y-weighting of grid cell in which column is located
             wcx = (1.0-abs(ris))    ! 1.0 ... 0.5
             wcy = (1.0-abs(rjs))    ! 1.0 ... 0.5


             !=================================================================
             ! if index of neighbour is exceeding range of region set 
             ! neighbour = current cell (i.e. no interpolation)
             ! in case of cyclic x-boundaries take corresponding cyclic i index
             !=================================================================
             
             if ( jsn == 0) jsn=1
             if ( jsn == jm(region)+1 ) jsn=jm(region)  ! isn-->jsn (wouter Peters)
             if ( xcyc(region) == 0 ) then
                ! non-cyclic boundaries
                if ( isn == 0) isn=1
                if ( isn == im(region)+1 ) isn=im(region)
             else
                ! cyclic x-boundaries
                if ( isn == 0 ) isn=im(region)
                if ( isn == im(region)+1 ) isn=1
             end if

             !============================================
             ! interpolate the gph to xy position of column
             !============================================
             !ls = 1   !layer
             do l=0,lm(region)
                !CMK note: gph now from 1-->lm+1 (dec 2002)
                height(l) =     wcx *      wcy*  gph(is,js,l+1)  + &
                           (1.0-wcx)*      wcy*  gph(isn,js,l+1) + &
                                wcx *(1.0-wcy)*  gph(is,jsn,l+1) + &
                           (1.0-wcx)*(1.0-wcy)*  gph(isn,jsn,l+1)
             end do

             !===========================
             ! determine layer of column
             !===========================
             do l=0,lm(region)
                if(height(l) .gt. stat(i_stat)%height) exit
             end do

             select case(l)
             case(0)
                ! force column to model surface
                ls = 1
                rls = -0.5  
             case default
                ls = l       ! column's layer
                ! the off-set from the center of the layer (-0.5--->+0.5) 
                ! (interpolation is in (m))
                rls = (stat(i_stat)%height - height(l-1)) / &
                     (height(l)-height(l-1)) - 0.5
             end select
             stat(i_stat)%ls = ls  ! store for output to file

             rls = 0.0 ! no interpolation
             wcz = 1.0 ! no interpolation

             !============================================
             ! if not yet done, interpolate the orography
             ! to xy position of column
             !============================================
             if (stat(i_stat)%oro < -10000) then
                 stat(i_stat)%oro =     wcx *      wcy*  oro(is,js)  + &
                           (1.0-wcx)*      wcy*  oro(isn,js) + &
                                wcx *(1.0-wcy)*  oro(is,jsn) + &
                           (1.0-wcx)*(1.0-wcy)*  oro(isn,jsn)
             endif

             
             do ispec=1,column_nspec
                n = column_speclist(ispec)   
                lst = ls 
                lstn = lsn 
                !========================================
                ! value of grid box without interpolation
                !========================================
                vmr = rm(is,js,:,n) / m(is,js,:) *  fscale(n)
                stat(i_stat)%c_mean(ispec,1,:) = &
                   stat(i_stat)%c_mean(ispec,1,:) + vmr
                stat(i_stat)%c_std(ispec,1,:) = &
                   stat(i_stat)%c_std(ispec,1,:) + vmr*vmr
                !========================================
                ! interpolation using slopes  
                !========================================
                !rm-value is obtained from rm + slopes. 
                !slope = rxm = (rm*dX/dx *deltaX/2)
                rm_interpol = rm(is,js,:,n) + 2.0*(ris*rxm(is,js,:,n) + &
                     rjs*rym(is,js,:,n) + &
                     rls*rzm(is,js,:,n) )
                rm_interpol = max(0.0,rm_interpol) 
                vmr = rm_interpol/m(is,js,:) *fscale(n)
                stat(i_stat)%c_mean(ispec,2,:) = &
                   stat(i_stat)%c_mean(ispec,2,:) + vmr
                stat(i_stat)%c_std(ispec,2,:) = &
                   stat(i_stat)%c_std(ispec,2,:) + vmr*vmr

                !========================================
                ! 3D linear interpolation 
                !========================================
                ! this PE handles also the neighbour layer
!PB             if ( lstn >= 1 .and. lstn <= lmr) then
!PBi
                rm_interpol = wcx      *wcy          *rm(is,js,:,n)   /m(is,js,:)   + & 
                                 (1.0-wcx)*wcy          *rm(isn,js,:,n)  /m(isn,js,:)  + &
                                 wcx      *(1.0-wcy)    *rm(is,jsn,:,n)  /m(is,jsn,:)  + &
                                 (1.0-wcx)*(1.0-wcy)    *rm(isn,jsn,:,n) /m(isn,jsn,:) 

                vmr = rm_interpol*fscale(n)
                stat(i_stat)%c_mean(ispec,3,:) = &
                   stat(i_stat)%c_mean(ispec,3,:) + vmr
                stat(i_stat)%c_std(ispec,3,:) = &
                   stat(i_stat)%c_std(ispec,3,:) + vmr*vmr


             end do  !ispec



                !======================
                !meteo data at columns
                !======================
                if (output_columns_meteo) then              

                   !============================================
                   ! gph already interpolated to xy position of column
                   !============================================
                   stat(i_stat)%metdat(8,:) = &
                        stat(i_stat)%metdat(8,:) + height

                   !============================================
                   ! interpolate the blh to xy position of column
                   !============================================
                   blh_interpol =     wcx *      wcy*  blh(is,js)  + &
                                     (1.0-wcx)*      wcy*  blh(isn,js) + &
                                     wcx *(1.0-wcy)*  blh(is,jsn) + &
                                     (1.0-wcx)*(1.0-wcy)*  blh(isn,jsn)
                   stat(i_stat)%metdat(7,:) = &
                      stat(i_stat)%metdat(7,:) + blh_interpol
                   !====================================
                   ! 3D linear interpolation Temperature
                   !====================================
                   T_interpol = &
                        wcx      *wcy      *wcz       *T(is,js,:)    + & 
                        (1.0-wcx)*wcy      *wcz       *T(isn,js,:)   + &
                        wcx      *(1.0-wcy)*wcz       *T(is,jsn,:)   + &
                        (1.0-wcx)*(1.0-wcy)*wcz       *T(isn,jsn,:) 

                   stat(i_stat)%metdat(1,:) = &
                      stat(i_stat)%metdat(1,:) + T_interpol
                   !====================================
                   ! 3D linear interpolation pressure
                   !====================================
                   p_interpol =(((0.5-rls) * phlb(is,js,1:lm(1))  + (0.5+rls) * phlb(is,js,2:lm(1)+1))  * wcx         * wcy       + &
                        ((0.5-rls) * phlb(isn,js,1:lm(1)) + (0.5+rls) * phlb(isn,js,2:lm(1)+1)) * (1.0-wcx)   * wcy       + &
                        ((0.5-rls) * phlb(is,jsn,1:lm(1)) + (0.5+rls) * phlb(is,jsn,2:lm(1)+1)) * wcx         * (1.0-wcy) + &
                        ((0.5-rls) * phlb(isn,jsn,1:lm(1))+ (0.5+rls) * phlb(isn,jsn,2:lm(1)+1))* (1.0-wcx)   * (1.0-wcy) ) *0.01 ![Pa] -> [hPa]

                   stat(i_stat)%metdat(2,:) = &
                      stat(i_stat)%metdat(2,:) + p_interpol

                   !====================================
                   ! 3D windspeed_u_interpol (x-direction)
                   !====================================

                   ! pu is eastward flux through east grid box boundary
                   ! windspeed_u_interpol wind component from east     

                   ! latitude of grid cell center js  [degrees]
                   lat_j =ybeg(region)+(js -0.5)*dyr
                   ! latitude of grid cell center jsn [degrees]
                   lat_jn=ybeg(region)+(jsn-0.5)*dyr
                   ! x-extension of grid cell js   [m]
                   ddx_j =ae * twopi * dxr / 360.0 * cos(lat_j*twopi/360.0)
                   ! x-extension of grid cell jsn  [m]
                   ddx_jn=ae * twopi * dxr / 360.0 * cos(lat_jn*twopi/360.0)

                   is_west=is-1
                   if ( xcyc(region) == 0 ) then
                      ! non-cyclic boundaries
                      if ( is_west == 0 ) is_west=1
                   else
                      ! cyclic x-boundaries
                      if ( is_west == 0 ) is_west=im(region)
                   end if

                   !west border                         !east border  
                   windspeed_u_interpol=-(((0.5-ris) * pu(is_west,js,:)/m(is_west,js,:)                                   + &
                        (0.5+ris) * pu(is,js,:)/ m(is,js,:) )      * &
                        ddx_j  *      wcy *      wcz          + &  
                        ((0.5-ris) * pu(is_west,jsn,:)/m(is_west,jsn,:)                                 + & 
                        (0.5+ris) * pu(is,jsn,:)/m(is,jsn,:) )     * &
                        ddx_jn * (1.0-wcy)*      wcz          + &  
                        ((0.5-ris) * pu(is_west,js,:)/m(is_west,js,:)                                 + &
                        (0.5+ris) * pu(is,js,:)/ m(is,js,:))    * &
                        ddx_j  *      wcy * (1.0-wcz)         + &  
                        ((0.5-ris) * pu(is_west,jsn,:)/m(is_west,js,:)                                + &
                        (0.5+ris) * pu(is,jsn,:)/ m(is,jsn,:) ) * &
                        ddx_jn * (1.0-wcy)* (1.0-wcz)         )
                   !====================================
                   ! 3D windspeed_v_interpol (y-direction)
                   !====================================

                   ! pv northward flux through south grid box boundary 
                   ! windspeed_v_interpol wind component from north    

                   ddy =ae * twopi * dyr / 360.0 ! y-extension of grid cell [m]

                   js_south = js-1
                   if (js_south==0) js_south=1

                   js_north = js+1
                   if (js_north==(jm(region)+1)) js_north=jm(region)

                   !south border                  !north border  
                   windspeed_v_interpol=-(( &
                        (0.5-rjs)*pv(is,js,:)/m(is,js_south,:)       + &
                        (0.5+rjs)*pv(is,js_north,:)/ m(is,js,:) )     * &
                        wcx *      wcz           + &  
                        ((0.5-rjs)*pv(isn,js,:)/m(isn,js_south,:)    + &
                        (0.5+rjs)*pv(isn,js_north,:)/ m(isn,js,:) )   * &
                        (1.0 -  wcx)*      wcz           + &     
                        ((0.5-rjs)*pv(is,js,:)/m(is,js_south,:)    + &
                        (0.5+rjs)*pv(is,js_north,:)/ m(is,js,:) )   * &
                        wcx * (1.0-wcz)          + &  
                        ((0.5-rjs)*pv(isn,js,:)/m(isn,js_south,:)  + &
                        (0.5+rjs)*pv(isn,js_north,:)/ m(isn,js,:) ) * &
                        (1.0 -  wcx)* (1.0-wcz)        ) * &
                        ddy 


                   windspeed_interpol = &
                        sqrt(windspeed_u_interpol**2 + windspeed_v_interpol**2)

                   stat(i_stat)%metdat(3,:) = &
                      stat(i_stat)%metdat(3,:) + windspeed_u_interpol
                   stat(i_stat)%metdat(4,:) = &
                      stat(i_stat)%metdat(4,:) + windspeed_v_interpol
                   stat(i_stat)%metdat(5,:) = &
                      stat(i_stat)%metdat(5,:) + windspeed_interpol

                endif  ! meteo out...
                stat(i_stat)%ncount = stat(i_stat)%ncount + 1
           end if
       end do ! end i_stat loop

    endif  ! myid = root_t

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(gph)
    nullify(oro)
    nullify(t)
    nullify(pu)
    nullify(pv)
    nullify(phlb)
    nullify(blh)
#ifdef MPI    
    deallocate(rmg)
    deallocate(rxmg)
    deallocate(rymg)
    deallocate(rzmg)
    deallocate(mg)
#endif

  end subroutine output_columnconc



  subroutine reset_columnconc_accumulator

    !==================================================================
    ! initialize accumulation arrays for average column concentration
    !==================================================================
    use ParTools, only: root_t, myid

    implicit none

    integer i_stat

    if (myid /= root_t) return

    do i_stat=1, n_stat
       stat(i_stat)%c_mean(:,:,:) = 0.0
       stat(i_stat)%c_std(:,:,:)  = 0.0
       stat(i_stat)%metdat(:,:)   = 0.0
       stat(i_stat)%ncount  = 0
    end do

  end subroutine reset_columnconc_accumulator
!
!
!
  subroutine evaluate_columnconc
    use dims, only  : idate
    use ParTools, only: myid, root
    use file_hdf,    only : Init, WriteData, TSds, Done, select
    use file_hdf,    only : CheckInfo
    use toolbox,     only : escape_tm
    !==================================================================
    ! evaluate mean concentration and std
    !==================================================================

    implicit none

    integer i_stat, i, ii,iii, io_status
    integer n_data, isds, nsds, itp
    real temp
    character(len=12)                    :: xname
    real, dimension(:,:), allocatable    :: cout                 ! to get output
    integer, dimension(:,:), allocatable :: iidate
    real                                 :: windspeed_u_interpol ! wind speed [m/s] x-direction , interpolated to column location
    real                                 :: windspeed_v_interpol ! wind speed [m/s] y-direction , interpolated to column location
    real                                 :: windspeed_interpol   ! wind speed [m/s]             , interpolated to column location
    real                                 :: winddir_interpol     ! wind direction               , interpolated to column location 
    if(myid /= root) return

    do i_stat=1, n_stat
       n_data = stat(i_stat)%ncount
       if (n_data > 0) then
          stat(i_stat)%c_mean(:,:,:) = stat(i_stat)%c_mean(:,:,:) / n_data
          stat(i_stat)%metdat(:,:)   = stat(i_stat)%metdat(:,:)   / n_data
          if (n_data > 1) then
             do i=1, column_nspec
                do ii= 1, n_interpol
                do iii= 1, lm(1)
                   temp = ((stat(i_stat)%c_std(i,ii,iii) -  n_data * ((stat(i_stat)%c_mean(i,ii,iii))**2) ) / (n_data-1) )
                   if(temp > 0) then 
                      stat(i_stat)%c_std(i,ii,iii) = sqrt(temp)
                   else
                      stat(i_stat)%c_std(i,ii,iii) = 0.0
                   endif
                enddo
                enddo
             enddo
          else
             stat(i_stat)%c_std(:,:,:)=-1.0
          end if
       else
          stat(i_stat)%c_mean(:,:,:)= -1.0
          stat(i_stat)%c_std(:,:,:) = -1.0
          stat(i_stat)%metdat(:,:)  = -999.
       end if

       do i=1,lm(1)
          windspeed_u_interpol = stat(i_stat)%metdat(3,i)
          windspeed_v_interpol = stat(i_stat)%metdat(4,i)
          windspeed_interpol   = stat(i_stat)%metdat(5,i)

          !====================================
          ! wind direction
          !====================================

          ! definition of winddirection:          
          ! east     :  90
          ! south    : 180
          ! west     : 270
          ! north    : 360
          if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol > 0))then              ! wind from north east (0...90)
             winddir_interpol = atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
          else if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol < 0)) then     ! wind from south east (90...180)
             winddir_interpol = 180.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
          else if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol == 0)) then    ! wind from east (90)
             winddir_interpol = 90.0
          else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol > 0)) then     ! wind from north west (270... 360)
             winddir_interpol = 360.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
          else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol < 0)) then     ! wind from south west (180...270)
             winddir_interpol = 180.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
          else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol == 0)) then    ! wind from west (270)
             winddir_interpol = 270.0
          else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol > 0)) then    ! wind from north (360)
             winddir_interpol = 360.0
          else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol < 0)) then    ! wind from south (180)
             winddir_interpol = 180.0
          else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol == 0)) then   ! no wind
             winddir_interpol = -1.0
          else
             call escape_tm('output_columnconc: error wind direction')
          end if
          stat(i_stat)%metdat(6,i) = winddir_interpol
        end do
    end do
  end subroutine evaluate_columnconc

  subroutine write_columnconc
    use dims, only  : idate
    use ParTools, only: myid, root
    use file_hdf,    only : Init, WriteData, TSds, Done
    use file_hdf,    only : WriteAttribute
    use toolbox,     only : escape_tm
    !==================================================================
    ! write mean concentration
    !==================================================================
    implicit none

    integer i, ii, io_status, i_stat
    integer n_data, isds, nsds, itp
    real, dimension(:,:,:), allocatable    :: cout                 ! to get output
    integer, dimension(:,:), allocatable :: iidate

    if(myid /= root) return
    ! output
    if (io_record == 0) then
       call WriteAttribute( io_hdf, 'column_ls', stat(:)%ls,cstatus)
       call WriteAttribute( io_hdf, 'column_oro', stat(:)%oro,cstatus)
    endif

    allocate(cout(n_stat,lm(1),1))

    do i=1, column_nspec
       do ii= 1,n_interpol
          do i_stat = 1, n_stat
            cout(i_stat,:,1) = stat(i_stat)%c_mean(i,ii,:)
          enddo
          call writedata(sds(i,ii), cout,  cstatus,start=(/0,0,io_record/))
       enddo
    enddo
    do i=1, n_metdat
       do i_stat = 1, n_stat
          cout(i_stat,:,1) = stat(i_stat)%metdat(i,:)
       enddo
       call writedata(sds_meteo(i), cout, cstatus, start=(/0,0,io_record/))
    enddo
    deallocate(cout)
    allocate(iidate(6,1))
    iidate(:,1) = idate
    call writedata(sds_idate, iidate, cstatus, start=(/0,io_record/))
    deallocate(iidate)
    io_record = io_record + 1
  end subroutine write_columnconc


end module user_output_column
